/*
**   IRC - Internet Relay Chat, source/commands.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   Portions Copyright (C) 2001 Adam Loghry
** 
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"
// #include "cmds.h"
#include "../lib/confparse/configparse.h"

// RCSID("$Id: commands.c,v 1.20 2001/10/02 01:40:35 cbehrens Exp $");

#ifdef TS5
#define TS_CURRENT  5   /* current TS protocol version */
#else
#define TS_CURRENT  3   /* current TS protocol version */
#endif

#ifdef TS5_ONLY
#define TS_MIN      5   /* minimum supported TS protocol version */
#else
#define TS_MIN      3   /* minimum supported TS protocol version */
#endif


char *months[] =
{
        "January",
        "February",
        "March",
        "April",
    "May",
        "June",
        "July",
        "August",
    "September",
        "October",
        "November",
        "December"
};

char *weekdays[] =
{
        "Sunday",
        "Monday",
        "Tuesday",
        "Wednesday",
    "Thursday",
        "Friday",
        "Saturday"
};

typedef struct _ircd_mode               IRCD_MODE;
typedef struct _ircd_capab              IRCD_CAPAB;

struct _ircd_mode
{
        char                            mode;
        char                            plusminus;
        IRCD_CLIENT                     *cptr;    
       
        char                            *banstring;
        char                            *exceptstring;  /* all three of these variables could probably share the same space, it might */
        char                            *quietstring;   /* be worth coming back and setting it up that way ...... later */              
       
        IRCD_MODE                       *next;
};                                            

struct _ircd_capab
{
        int                                     value;
        char                            *name;        
        int                                     flags;
};     

int _register_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr);
char *find_wlvl_text(int number);                        
int m_message(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args, int not);
int _do_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *username, char *hostname, char *servername, char *ircname);
int _lusers_send(IRCD_CLIENT *pptr, char *pattern, int sendcounter);                                                
int _motd_send(IRCD_CLIENT *pptr);
void _sendnick(IRCD_CLIENT *cptr, IRCD_CLIENT *acptr);
void _build_channel_modes(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr, char *mbuf, char *pbuf);
void _send_channel_modes(IRCD_CLIENT *cptr, IRCD_CHANNEL *chptr);                         
void _umode_build_change(IRCD_CLIENT *pptr, char *buf, int oldmodes, int sendmask);
void _umode_build(IRCD_CLIENT *pptr, char *buf, int mask);                         
int _channel_canjoin(IRCD_CLIENT *pptr, IRCD_CHANNEL *chptr, char *key);
IRCD_CHANNELLIST *_channel_ismember(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr);
void _channel_add_user(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, int flag);   
void _channel_del_user(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, IRCD_CHANNELLIST *chlptr);
int _channel_cansend(IRCD_CLIENT *pptr, IRCD_CHANNEL *chptr);                            
void _channel_invite_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr);
#ifdef HMC
int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *target, char *prefix, char *args, char *hmc);
#else
int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *target, char *prefix, char *args);
#endif
IRCD_MODE *_ircdmode_create(IRCD_MODE **ircdmodes, IRCD_MODE **lastircdmode);
void _ircdmode_free(IRCD_MODE **bucket);
int _channel_change_opflag(IRCD_CHANNEL *chptr, IRCD_CLIENTLIST *clptr, IRCD_MODE *opmodeptr, int mine);
char *_make_nick_user_host(char *nick, char *user, char *host, char *buf, char **hostptr);
int _channel_ban_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *str, char *hostptr, ircd_uint32_t net, ircd_uint32_t mask);
int _channel_ban_del(IRCD_CHANNEL *chptr, char *str);
int _channel_except_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *str, char *hostptr, ircd_uint32_t net, ircd_uint32_t mask);
int _channel_except_del(IRCD_CHANNEL *chptr, char *str);
int _channel_quiet_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *str, char *hostptr, ircd_uint32_t net, ircd_uint32_t mask);
int _channel_quiet_del(IRCD_CHANNEL *chptr, char *str);
void do_who(IRCD_CLIENT *pptr, IRCD_CLIENT *acptr, IRCD_CHANNEL *chptr, IRCD_CLIENTLIST *clptr);
char *datestr(int thetime, char *buf);
#ifdef KLINES
char *_reason_fix(char *reason);
char *_cluster(char *hostname, int *ip_or_host);
#endif
void _send_var(IRCD_CLIENT *pptr, IRCD_VAR *vptr);
void _send_var_to_opers(IRCD_CLIENT *pptr, IRCD_VAR *vptr);
void _send_servers_recurs(IRCD_CLIENT *lptr, IRCD_CLIENT *cptr);
void _count_servers_and_users(IRCD_CLIENT *acptr, int *numservers, int *numusers);
void send_capabilities(IRCD_CLIENT *cptr, int send_zip);
void _channelname_clean(unsigned char *chname);
int m_useriphost(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args, int justip);
int do_netset(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *var, char *value);

int _user_modes[] =
{
	CLIENT_UMODES_OPER,		'o',
	CLIENT_UMODES_LOCOP,		'O',
	CLIENT_UMODES_FMODE,		'f',
	CLIENT_UMODES_IMODE,		'i',
	CLIENT_UMODES_KMODE,		'k',
	CLIENT_UMODES_SMODE,		's',
	CLIENT_UMODES_WMODE,		'w',
	CLIENT_UMODES_ZMODE,		'z',
	CLIENT_UMODES_YMODE,		'y',
	CLIENT_UMODES_CMODE,		'c',
	CLIENT_UMODES_NMODE,		'n',
	CLIENT_UMODES_SERVICE,		'x',
	CLIENT_UMODES_REGISTERED,	'r',
	CLIENT_UMODES_ONLYREGMSGS,	'R',
	CLIENT_UMODES_DEAF,		'd',
	0, 0
};

int _channel_modes[] =
{
	CHANNEL_MODES_PRIVATE,		'p',
	CHANNEL_MODES_SECRET,		's',
	CHANNEL_MODES_MODERATED,	'm',
	CHANNEL_MODES_NOPRIVMSGS,	'n',
	CHANNEL_MODES_TOPICLIMIT,	't',
	CHANNEL_MODES_INVITEONLY,	'i',
	CHANNEL_MODES_OPERONLY,		'O', /* adopted from DALnet */
	CHANNEL_MODES_REGGEDONLY,	'R', /* also adopted from there, only few modes i liked */
	0, 0
};

IRCD_USERFLAG _user_flags[] =
{
	{ CLIENT_NFLAGS_OWALLOPS,			"OWALLOPS",			1	},
	{ CLIENT_NFLAGS_SWALLOPS,			"SWALLOPS",			1	},
	{ CLIENT_NFLAGS_STATSNOTICES,			"STATSNOTICES",			1	},
	{ CLIENT_NFLAGS_OKILLS,				"OKILLS",			1	},
	{ CLIENT_NFLAGS_SKILLS,				"SKILLS",			1	},
	{ CLIENT_NFLAGS_SNOTICES,			"SNOTICES",			0	},
	{ CLIENT_NFLAGS_CLICONNECTS,			"CLICONNECTS",			1	},
	{ CLIENT_NFLAGS_CLIDISCONNECTS,			"CLIDISCONNECTS",		1	},
	{ CLIENT_NFLAGS_THROTTLES,			"THROTTLES",			1	},
	{ CLIENT_NFLAGS_NICKCHANGES,			"NICKCHANGES",			1	},
	{ CLIENT_NFLAGS_IPMISMATCHES,			"IPMISMATCHES", 		1	},
	{ CLIENT_NFLAGS_LWALLOPS,			"LWALLOPS",			1	},
	{ CLIENT_NFLAGS_CONNECTS,			"CONNECTS",			1	},
	{ CLIENT_NFLAGS_SQUITS,				"SQUITS",			1	},
	{ CLIENT_NFLAGS_GLINEEXPIRES,			"EXPIREDGLINES",		1,	},
#if 0
	{ CLIENT_NFLAGS_GLINES,			"GLINES",		1, 	}, 
#endif
	{ CLIENT_NFLAGS_INVISIBLE,		"INVISIBLE",		0	},
	{ CLIENT_NFLAGS_DEAF,			"DEAF",			0	},
	{ CLIENT_NFLAGS_ONLYREGISTEREDMSGS,	"ONLYREGISTEREDMSGS", 	0	},
	{ 0, NULL, 0 }
};

#ifdef HMC
IRCD_HMCLASS _hostmask_classes[] =
{
	{ "R",	}, /* registered */
	{ "U",	}, /* normaluser */
	{ "D",	}, /* deaf guy */
	{ "O",	}, /* operator(global && local) */
	{ NULL }
};
#endif
IRCD_CAPAB _ircd_capabs[] =
{
#if 0
	{ IRCD_CAPAB_VALUE_IPADDR,		"IPADDR",	0 },
	{ IRCD_CAPAB_VALUE_ZIP,			"ZIP",		0 },
	{ IRCD_CAPAB_VALUE_EMODE,		"EX",		0 },
#endif
	{ IRCD_CAPAB_VALUE_NOQUITSTORM,	"QS",		0 },
	{ 0, NULL, 0 }
};

int hostname_check(char *hostname)
{
	char *tptr = hostname;
	char *dotptr = NULL;

	for(;*tptr;tptr++)
	{
		if ((*tptr <= ' ') || (*tptr > 'z') || (*tptr == ':') ||
				(*tptr == '*') || (*tptr == '%'))
			break;
		if (*tptr == '.')
			dotptr = tptr;
	}
	if (*tptr || !*hostname || !dotptr || (strlen(dotptr) > 4))
		return 0;

	return 1;
}

int nickname_check(char *nick)
{
	char	*ch;

	if ((*nick == '-') || isdigit(*nick))
		return 0;
	for (ch = nick; *ch && (ch - nick) < NICKLEN; ch++)
		if (!isvalid(*ch) || isspace(*ch))
			break;
	*ch = '\0';
	return (ch - nick);
}
